/**
 * JASMINe
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbstractDroolsWorkingMemory.java 4708 2009-08-04 13:56:37Z nicoulaj $
 * --------------------------------------------------------------------------
 */

package org.ow2.jasmine.rules;

import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.drools.FactException;
import org.drools.FactHandle;
import org.drools.RuleBase;
import org.drools.WorkingMemory;
import org.drools.compiler.PackageBuilder;
import org.drools.compiler.PackageBuilderConfiguration;
import org.drools.rule.builder.dialect.java.JavaDialectConfiguration;
import org.ow2.jasmine.rules.api.DroolsWorkingMemoryState;
import org.ow2.jasmine.rules.api.IDroolsWorkingMemory;

public abstract class AbstractDroolsWorkingMemory implements
        IDroolsWorkingMemory {

    /**
     * Logger.
     */
    private static Logger logger = null;

    /**
     * Thread to fire rules.
     */
    protected Thread thread = null;

    /**
     * Period to fire rules.
     */
    private volatile long timeout = 5000;

    /**
     * A directory to save generated files.
     */
    protected File dumpDir;

    /**
     * The managed working memory.
     */
    protected volatile WorkingMemory workingMemory = null;

    /**
     * {@link ClassLoader} to build rules.
     */
    protected final ClassLoader classLoader;

    /**
     * The RuleBase
     */
    protected RuleBase ruleBase;

    /**
     * The state of the working memory
     */
    protected volatile DroolsWorkingMemoryState state;

    /**
     * The configuration used to create package builders
     */
    protected final PackageBuilderConfiguration pkgBuilderCfg;
    
    
    // ------------------------------------------------------------------------
    // Constructor
    // ------------------------------------------------------------------------

    public AbstractDroolsWorkingMemory(final ClassLoader classloader,
    								   final Logger logger) {
    	
        classLoader = classloader;
        AbstractDroolsWorkingMemory.logger = logger;
        state = DroolsWorkingMemoryState.IDLE;
        this.pkgBuilderCfg = new PackageBuilderConfiguration(classLoader);
        JavaDialectConfiguration javaConf = (JavaDialectConfiguration) pkgBuilderCfg.getDialectConfiguration("java");
        javaConf.setCompiler(JavaDialectConfiguration.JANINO);
        if (dumpDir != null) {
            pkgBuilderCfg.setDumpDir(dumpDir);
        }
    }

    
    // ------------------------------------------------------------------------
    // Implementation of the IDroolsWorkingMemory interface
    // ------------------------------------------------------------------------

    /**
     * Stop the WorkingMemory, with the state STOPPED.
     * @see {@link AbstractDroolsWorkingMemory#stop(DroolsWorkingMemoryState)}
     */
    /* (non-Javadoc)
     * @see org.ow2.jasmine.rules.api.IDroolsWorkingMemory#stop()
     */
    public void stop() {
    	stop(DroolsWorkingMemoryState.STOPPED);
    }
    
    /**
     * Stop the working memory with a given state (allowed states are
     * STOPPED and IDLE)
     */
    public void stop(DroolsWorkingMemoryState state) {
        if (workingMemory != null) {
            workingMemory.halt();
        }
        if (thread != null) {
            thread.interrupt();
            thread = null;
        }
        workingMemory = null;
    	if( state == DroolsWorkingMemoryState.IDLE
    	 || state == DroolsWorkingMemoryState.STOPPED ) {
        	this.state = state;
    	} else {
    		this.state = DroolsWorkingMemoryState.STOPPED;
    	}
    }
    
    /**
     * Insert a fact in the WorkingMemory
     * @param fact the fact object to insert
     * @return the corresponding FactHandle
     */
    public FactHandle insert(final Object fact) {
        if (state != DroolsWorkingMemoryState.STOPPED && workingMemory != null) {
            return workingMemory.insert(fact);
        } else {
            return null;
        }
    }
    

    // ------------------------------------------------------------------------
    //
    // ------------------------------------------------------------------------

    /**
     * Set a directory to save generated files.
     *
     * @param dumpDir
     *            directory to save generated files
     */
    public void setDumpDir(final File dumpDir) {
        this.dumpDir = dumpDir;
    }

    // ------------------------------------------------------------------------
    // Protected methods
    // ------------------------------------------------------------------------

    /**
     * Create a new working memory for a given rule base.
     *
     * @param ruleBase
     *            rule base
     */
    protected void createWorkingMemory() {
        logger.log(Level.INFO, "Init the Working Memory...");
        workingMemory = ruleBase.newStatefulSession();
    }

    /**
     * Start the rule execution process
     */
    protected void startRuleExecutor() {
    	state = DroolsWorkingMemoryState.RUNNING;
        thread = new Thread(new RuleExecutor());
        thread.start();
    }

    /**
     * Periodically fire rules.
     */
    private class RuleExecutor implements Runnable {
        /**
         * Execute the rules after the timeout
         */
        public void run() {
            Logger logger = Logger.getLogger(this.getClass().getName());
            while (state.equals(DroolsWorkingMemoryState.RUNNING)) {
                try {
                    Thread.sleep(timeout);
                } catch (InterruptedException e) {
                    logger.log(Level.INFO, "DroolsWorkingMemory stopped");
                }

                if (workingMemory != null) {
                    try {
                        logger.log(Level.FINE, "Execute the rules...");
                        workingMemory.fireAllRules();
                    } catch (FactException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }
    
    /**
     * Get the RuleBase
     * @return the RuleBase
     */
    public RuleBase getRuleBase() {
        return ruleBase;
    }
    
    /**
     * Get the current state of the working memory
     * @return the state
     * @see {@link org.ow2.jasmine.rules.api.DroolsWorkingMemoryState}
     */
    public DroolsWorkingMemoryState getState() {
    	return state;
    }
    
    /**
     * Get a PackageBuilder to compile a rule package externally
     * @return an initialized PackageBuilder
     * @throws Exception if it could not be built
     */
    public PackageBuilder getPackageBuilder() throws Exception {
    	return new PackageBuilder(pkgBuilderCfg);
    }
}
